@@ -50,6 +50,7 @@
 #include "syncdb.h"
 #include "option.h"
 #include "authize.h"
+#include "exitcodes.h"
 #include "abook.h"
 #include "abook_ldap.h"
 
@@ -942,12 +943,13 @@ abook_ldap_fetch(char *alias, int *count)
 	    fptr->field = strdup(secondaryconfig.map[i].field);
 	    if (secondaryconfig.map[i].append != NULL) {
 	      int fptrdatalen = strlen(values[0])+strlen(secondaryconfig.map[i].append)+1;
-	      printf("secondary appending %s to field: %s value: %s\n", secondaryconfig.map[i].append, secondaryconfig.map[i].field,values[0]);
 	      fptr->data=malloc(fptrdatalen);
-	      fptr->data=strlcpy(fptr->data, values[0], fptrdatalen);
-	      fptr->data=strlcat(fptr->data, secondaryconfig.map[i].append, fptrdatalen);
+	      if(!fptr->data) fatal("out of memory", EC_TEMPFAIL);
+	      strlcpy(fptr->data, values[0], fptrdatalen);
+	      strlcat(fptr->data, secondaryconfig.map[i].append, fptrdatalen);
 	    } else {
 	      fptr->data  = strdup(values[0]);
+	      if(!fptr->data) fatal("out of memory", EC_TEMPFAIL);
 	    }
 	    (*count)++;
 	    fptr++;
