@@ -1,3 +1,5 @@
+#ifndef __authize_h
+#define __authize_h
 /* authize.h -- definitions for authorization API
  *
  * Copyright (c) 1993-2000 Carnegie Mellon University.  All rights reserved.
@@ -129,4 +131,4 @@ char *auth_username(auth_id *id);
 /* return the access level for a given id
  */
 int auth_level(auth_id *id);
-
+#endif
