@@ -317,7 +317,7 @@ int bb_init()
 		    newval = malloc(size = len + 256);
 		    if (!newval) break;
 		}
-		sprintf(newval, "* %c (%s) ", mbox[-1], host);
+		snprintf(newval, size, "* %c (%s) ", mbox[-1], host);
 		response = bb_addacl(newdb, con[hnum], mbox, &newval, &size);
 		if (response) break;
 	    }
@@ -497,7 +497,7 @@ int bb_subsinit(user)
     char dbname[256];
 
     /* check if it already exists */
-    sprintf(dbname, "user/%s/subs", user);
+    snprintf(dbname, sizeof(dbname), "user/%s/subs", user);
 
     return (sdb_check(dbname) == 0 || sdb_create(dbname) == 0 ? 0 : -1);
 }
@@ -521,10 +521,10 @@ int bb_subscribe(id, name, subscribe)
     /* get old value / verify bboard name / verify proper rights */
     user = auth_username(id);
     if (!strcasecmp(inbox, name)) {
-	sprintf(iboxname, "user.%s", user);
+	snprintf(iboxname, sizeof(iboxname), "user.%s", user);
 	name = iboxname;
     }
-    sprintf(dbname, "user/%s/subs", user);
+    snprintf(dbname, sizeof(dbname), "user/%s/subs", user);
     if (sdb_writelock(dbname, name, SDB_ICASE) < 0) return (-1);
     if (sdb_get(dbname, name, SDB_ICASE, &value) < 0
 	|| bb_get(name, &bbvalue, NULL, &acl, NULL) < 0
@@ -583,7 +583,7 @@ int bb_matchstart(state, user, pat)
 
     /* initialize match state */
     memset((char *) state, 0, sizeof (bb_state));
-    sprintf(state->inboxname, "user.%s", user);
+    snprintf(state->inboxname, sizeof(state->inboxname), "user.%s", user);
     state->inboxlen = strlen(state->inboxname);
     state->lastmin = -1;
     end = pat + strlen(pat) - 1;
@@ -608,7 +608,7 @@ int bb_matchstart(state, user, pat)
 		  &state->bb, &bbcount) >= 0) {
 	state->bbend = state->bb + bbcount;
 	state->bbpos = state->bb;
-	sprintf(dbname, "user/%s/subs", user);
+	snprintf(dbname, sizeof(dbname), "user/%s/subs", user);
 	if (sdb_match(dbname, pat, GLOB_ICASE | GLOB_HIERARCHY, NULL,
 		      1, &state->subs, &subcount) >= 0) {
 	    state->subend = state->subs + subcount;
@@ -653,7 +653,7 @@ int bb_matchverify(state, id, pat, oldname, newname, autosub)
 
     /* set database name so we have it if needed */
     user = auth_username(id);
-    sprintf(dbname, "user/%s/subs", user);
+    snprintf(dbname, sizeof(dbname), "user/%s/subs", user);
 
     /* build bbsubpos array & check for deleted/renamed bboards */
     while (state->subpos < state->subend) {
@@ -770,7 +770,7 @@ char *bb_match(state, id, name, loc, sep_char, flags, subonly)
     /* check for inbox */
     *flags = 0;
     if (!state->inboxdone) {
-	sprintf(dbname, "user/%s/subs", auth_username(id));
+	snprintf(dbname, sizeof(dbname), "user/%s/subs", auth_username(id));
 	if (sdb_get(mboxdb, state->inboxname,
 		    SDB_ICASE | SDB_QUICK, &uid) >= 0
 	    && bb_get(NULL, &uid, loc, &acl, sep_char) == 0
@@ -974,7 +974,7 @@ char *bb_create(id, mbox, count, hostlist)
 	if (!value) {
 	    reply = rpl_nomem;
 	} else {
-	    sprintf(value, "* . (%s) ", host);
+	    snprintf(value, size, "* . (%s) ", host);
 	    reply = bb_addacl(mboxdb, con, mbox, &value, &size);
 	}
     }
@@ -1115,9 +1115,10 @@ char *bb_rename(id, oldname, newname, rflag)
     /* deal with inbox */
     if (!strcasecmp(newname, inbox)) return (rpl_inval);
     if (!strcasecmp(oldname, inbox)) {
-	iname = malloc(strlen(auth_username(id)) + 6);
+        int inamelen = strlen(auth_username(id)) + 6;
+	iname = malloc(inamelen);
 	if (!iname) return (rpl_nomem);
-	sprintf(iname, "user.%s", auth_username(id));
+	snprintf(iname, inamelen, "user.%s", auth_username(id));
 	oldname = iname;
     }
     
@@ -1216,7 +1217,7 @@ char *bb_rename(id, oldname, newname, rflag)
 	}
 
 	/* update user's subscriptions file */
-	sprintf(dbname, "user/%s/subs", auth_username(id));
+	snprintf(dbname, sizeof(dbname), "user/%s/subs", auth_username(id));
 	if (sdb_get(dbname, oldname, SDB_ICASE, &value) >= 0 && value != NULL) {
 	    bb_subscribe(id, newname, 1);
 	    bb_subscribe(id, oldname, 0);
@@ -1324,7 +1325,7 @@ int bb_last(mbox, uid, host)
 	    size = strlen(uid) + strlen(host) + 256;
 	    if ((con = imap_connect(host, NULL, 1)) != NULL
 		&& (new = malloc(size)) != NULL) {
-		sprintf(new, "%s . (%s) ", uid, host);
+		snprintf(new, size, "%s . (%s) ", uid, host);
 		if (bb_addacl(mboxdb, con, mbox, &new, &size) == NULL) {
 		    result = 0;
 		}
@@ -1365,10 +1366,10 @@ int bb_seen(mbox, uid, user)
 {
     char *value, *new = NULL;
     char dbname[256];
-    int result = -1, result1;
+    int result = -1, result1, newlen;
 
     /* set the db name */
-    sprintf(dbname, "user/%s/subs", user);
+    snprintf(dbname, sizeof(dbname), "user/%s/subs", user);
 
     /* sanity check for existance of mbox & subscription list */
     if (sdb_get(mboxdb, mbox, SDB_ICASE, &value) < 0 || !value
@@ -1381,9 +1382,10 @@ int bb_seen(mbox, uid, user)
     if (!value) value = "0";
 
     /* set new value */
-    new = malloc(strlen(uid) + 3);
+    newlen = strlen(uid) + 3;
+    new = malloc(newlen);
     if (new) {
-	sprintf(new, "%c %s", *value, uid);
+	snprintf(new, newlen, "%c %s", *value, uid);
 	result = sdb_set(dbname, mbox, SDB_ICASE, new);
     }
 
