@@ -1269,9 +1269,10 @@ static void imsp_getacl(fbuf, cp, tag, id, host)
 	    if (optnum == ACL_ADDRESSBOOK) {
 		acl = abook_getacl(id, item);
 		if ((acl != NULL) && (*acl == '\0')) {
-		    defacl = malloc(strlen(user) + 32);
+		    int defacllen = strlen(user) + 32;
+		    defacl = malloc(defacllen);
 		    if (defacl) {
-			sprintf(acl = defacl, "%s\t%s\t", user,
+			snprintf(acl = defacl, defacllen, "%s\t%s\t", user,
 				acl_masktostr(ACL_ALL, rbuf));
 		    }
 		}
@@ -1523,7 +1524,7 @@ void im_start(int fd, char *host)
 	if (p = strchr(buf, '\n')) *p = '\0';
 	/* can't have [ be first char, sigh */
 	for(p = buf; *p == '['; p++);
-	sprintf(tagbuf, msg_alert, p);
+	snprintf(tagbuf, sizeof(tagbuf), msg_alert, p);
 
 	if (write(fd, tagbuf, strlen(tagbuf)) < 0) {
 	  perror("write shutdown alert");
@@ -1533,7 +1534,7 @@ void im_start(int fd, char *host)
     }
 
     /* send greeting, setup idle autologout */
-    sprintf(tagbuf, msg_greeting, VERSION);
+    snprintf(tagbuf, sizeof(tagbuf), msg_greeting, VERSION);
     if (write(fd, tagbuf, strlen(tagbuf)) < 0) {
 	perror("write greeting");
 	close(fd);
@@ -1558,9 +1559,9 @@ void im_start(int fd, char *host)
 
     /* main protocol loop */
     while (dispatch_readline(fbuf) != NULL) {
-	/* get the tag */
+	/* get the tag - must not be NULL or greater than 64 characters long */
 	tag = get_atom(fbuf);
-	if (tag == (char *) NULL) {
+	if ((tag == (char *) NULL) || (strlen(tag) > 64)) {
 	    SEND_STRING(fbuf, msg_badtag);
 	} else {
 	    /* copy tag into tag-response buffer */
