@@ -51,7 +51,7 @@
 typedef struct option_state {
     /* all variables are private to option module */
     sdb_keyvalue *gkv, *ukv, *gkvpos, *ukvpos;
-    int gcount, ucount, buflen;
+    int gcount, ucount, gtotal, utotal, buflen;
     char *buf;
 } option_state;
 
