@@ -49,6 +49,7 @@
 #include <sasl/sasl.h>
 #include <netinet/in.h>
 
+#include "dispatch.h"
 #include "option.h"
 #include "syncdb.h"
 #include "util.h"
@@ -111,7 +112,7 @@ mysasl_make_secprops(void)
     prop->min_ssf = min;
     prop->max_ssf = max;
 
-    prop->maxbufsize = 4000;
+    prop->maxbufsize = MAX_BUF;
 
     prop->security_flags = 0;
 #ifndef ANONYMOUS_LOGIN
