@@ -61,6 +61,7 @@
 #include <sys/file.h>
 #include <sys/param.h>
 #include <fcntl.h>
+#include <errno.h>
 #include <syslog.h>
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
@@ -139,40 +140,40 @@ static void freecache(c)
     
     /* sanity checks */
     if (c == NULL) return;
-
+    
     if (c->kv != NULL) {
-
-      /* walk the cache element array freeing elements */
-      for (i = 0; i < c->cachecount; i++) {
-	if (c->kv[i].key != NULL) free(c->kv[i].key);
-	if (c->kv[i].value != NULL) free(c->kv[i].value);
-      }
-
-      /* free the cache array */
-      free((char *)c->kv);
-      c->kv = NULL;
+	
+	/* walk the cache element array freeing elements */
+	for (i = 0; i < c->cachecount; i++) {
+	    if (c->kv[i].key != NULL) free(c->kv[i].key);
+	    if (c->kv[i].value != NULL) free(c->kv[i].value);
+	}
+	
+	/* free the cache array */
+	free((char *)c->kv);
+	c->kv = NULL;
     }
-
+    
     /* reset cache counts to 0 */
     c->cachecount = 0;
     c->cachesize = 0;
-
+    
     /* reset cache state to unloaded */
     c->modified = 0;
     c->loaded = 0;
     if (c->fd != -1) {
-      if (c->fd < 3) {
-	if (c->fd != 0) { /* this stuff may be uninitialzed soo... */
-	  syslog(LOG_ERR, "imspd: not closing invalid fd of %d\n", c->fd);
+	if (c->fd < 3) {
+	    if (c->fd != 0) { /* this stuff may be uninitialzed soo... */
+		syslog(LOG_ERR, "imspd: not closing invalid fd of %d\n", c->fd);
+	    }
+	} else {
+	    close(c->fd);
 	}
-      } else {
-	close(c->fd);
-      }
     }
     c->fd = -1;
     if (c->locks > 0) {
-      fprintf(stderr,"Shouldn't free cache %s - %d locks still active\n", 
-	      c->db, c->locks);
+	fprintf(stderr,"Shouldn't free cache %s - %d locks still active\n", 
+		c->db, c->locks);
     }
 }
 
@@ -208,19 +209,19 @@ static cache *findcache(db)
     
 /* : try to load a private cache database */
     if (!strncmp(db, PRIVPREFIX, PRIVPREFIXLEN) && db[PRIVPREFIXLEN] == '/') {
-
+	
 /* : - pick off the private name */
 	scan = strchr(db + PRIVPREFIXLEN + 1, '/');
 	if (scan == NULL) return (NULL);
 	++scan;
-
+	
 /* : - walk the private database table matching the name */
 	for (i = 0; privdbstr[i]; ++i) {
-
+	    
 	    if (!strcmp(scan, privdbstr[i]) ||
 		(i >= PDBPREFIXPOS
 		 && !strncmp(scan, privdbstr[i], strlen(privdbstr[i])))) {
-
+		
 /* : -- if this is not the correct instance of the private database then flush it and get the right one */
 		if (strcmp(privdb[i].db + PREFIXLEN + 1, db)) {
 		    if (privdb[i].locks > 0) return (NULL);
@@ -271,9 +272,9 @@ static int ikeycmp(kv1, kv2)
  * END HISTORY */
 
 static int parsecache(c, data, flags)
-  cache *c;				/* U: cache to parse into */
-  char* data;				/* I: database data buffer */
-  int flags;				/* I: case insensitive if non-zero */
+    cache *c;				/* U: cache to parse into */
+    char* data;				/* I: database data buffer */
+    int flags;				/* I: case insensitive if non-zero */
 {
     int sorted;				/* source file was sorted if 1 */
     long lines;				/* number of lines in data */
@@ -284,26 +285,26 @@ static int parsecache(c, data, flags)
     char savech;			/* temporary save character */
     sdb_keyvalue *kv;			/* current keyvalue in cache */
     int (*cmpf)();			/* sort comparison function */
-
+    
 /* : initialization */
     lines = 0;
     sorted = 1;
     cmpf = (flags & SDB_ICASE) ? strcasecmp : strcmp;
-
+    
 /* : count the number of lines in the database data */
     lines = 0;
     for (scan = data; *scan != '\0'; scan++) {
 	if (*scan == '\n') lines++;
     }
-
+    
 /* : make sure that we count a line without a linefeed at the end */
     if (*(scan-1) != '\n') ++lines;
-
+    
 /* : CLAIM - the number of lines in the data is equivalent to the number of
-     key value pairs in the database.   We can base the size of the cache on
-     that value.   In fact, when we allocate the array, we allocate the required
-     number of slots plus a CACHE_INCREMENT that allows for growth of the cache
-     from new keyvalue pairs. */
+   key value pairs in the database.   We can base the size of the cache on
+   that value.   In fact, when we allocate the array, we allocate the required
+   number of slots plus a CACHE_INCREMENT that allows for growth of the cache
+   from new keyvalue pairs. */
 /* : allocate space for the cache array */
     size = ((lines + CACHE_INCREMENT) * sizeof(sdb_keyvalue));
     kv = (sdb_keyvalue *) malloc(size);
@@ -313,11 +314,11 @@ static int parsecache(c, data, flags)
     c->kv = kv;
     c->cachecount = lines;
     c->cachesize = lines + CACHE_INCREMENT;
-
+    
 /* : walk each line in the data parsing key value pairs */
     scan = data;
     for (kv = c->kv; lines; --lines, ++kv) {
-
+	
 /* : - parse the key, handling quoted characters */
 	token = scan;
 	for (dst = scan; (*scan != ' ') && (*scan != '\n') && (*scan != '\0'); scan++) {
@@ -332,20 +333,20 @@ static int parsecache(c, data, flags)
 	    }
 	    *dst++ = *scan;
 	}
-
+	
 /* : - save the current scan character and write a NUL terminator to the parsed string */
 	savech = *scan;
 	*dst = '\0';			/* this may overwrite *scan */
-
+	
 /* : - copy the parsed key to the keyvalue pair */
 	kv->key = strdup(token);
 	if (kv->key == NULL) {
 	    return(-1);
 	}
-
+	
 /* : - restore the saved scan character */
 	*scan = savech;
-
+	
 /* : - if at end of line or string then set the value to NULL */
 	if ((*scan == '\n') || (*scan == '\0')) {
 	    kv->value = NULL;
@@ -368,29 +369,29 @@ static int parsecache(c, data, flags)
 		*dst++ = *scan;
 	    }
 	    *dst = '\0';
-
+	    
 /* : - copy the parsed value to the keyvalue pair */
 	    kv->value = strdup(token);
 	    if (kv->value == NULL) {
 		return(-1);
 	    }
 	}
-
+	
 /* : - move scan to the start of the next line */
 	scan++;
-
+	
 /* : - check to make sure that the data is sorted */
 	if (sorted && (kv != c->kv) && ((*cmpf)(kv[-1].key, kv->key) > 0)) {
 	    sorted = 0;
 	}
     }
-
+    
 /* : sort it, only if necessary */
     if (!sorted) {
 	qsort(c->kv, c->cachecount, sizeof (sdb_keyvalue),
 	      (flags & SDB_ICASE) ? ikeycmp : keycmp);
     }
-
+    
 /* : return success */
     return(0);
 }
@@ -416,34 +417,34 @@ static int loadcache(c, flags)
     int count;				/* number of characters read from file */
     char* data;				/* raw (unparsed) database data */
     char lname[MAXDBPATHLEN + 5];	/* lock file name buffer */
-
+    
 /* : initialization */
     data = NULL;
     rtval = 0;
-
+    
 /* : quit if the cache is loaded and we don't care if it's stale */
     if (c->loaded && (flags & SDB_QUICK)) {
 	return (0);
     }
-
+    
 /* : quit if we can't stat the database file */
     if (stat(c->db, &stbuf) < 0) {
-      if (imspd_debug) {
-	fprintf(stderr,"%s: ", c->db);
-	perror("failed to stat cache buffer");
-      }
-      return(-1);
+	if (imspd_debug) {
+	    fprintf(stderr,"%s: ", c->db);
+	    perror("failed to stat cache buffer");
+	}
+	return(-1);
     }
-
+    
 /* : quit if the cache is up to date */
     if ((c->loaded == 1) && ((flags & SDB_ICASE) == c->icase)
 	&& (stbuf.st_mtime == c->mtime)) {
 	return(0);
     }
-
+    
 /* : free any existing cache structure */
     freecache(c);
-
+    
 /* : if the database file is 0 length then the refresh is complete */
     if (stbuf.st_size == 0) {
 	c->loaded = 1;
@@ -452,32 +453,32 @@ static int loadcache(c, flags)
 	c->cachecount = 0;
 	return(0);
     }
-
+    
 /* : open the database file if it hasn't already been opened */
     fd = c->fd;
     if (c->locks != 0 && (c->fd == -1)) {
-      if (imspd_debug) {
-	fprintf(stderr,"internal consistency error. We got locks but not a valid fd for %s\n",
-		c->db);
-	return(-1);
-      } else {
-	syslog(LOG_ERR,"internal consistency error. We got locks but not a valid fd for %s\n",
-	       c->db);
-      }
+	if (imspd_debug) {
+	    fprintf(stderr,"internal consistency error. We got locks but not a valid fd for %s\n",
+		    c->db);
+	    return(-1);
+	} else {
+	    syslog(LOG_ERR,"internal consistency error. We got locks but not a valid fd for %s\n",
+		   c->db);
+	}
     }
-
+    
     if ((c->locks == 0) || (c->fd == -1)) {
 	fd = open(c->db, O_RDONLY);
 	if (fd < 0) {
-	  if (imspd_debug) {
-	    fprintf(stderr,"%s: ", c->db);
-	    perror("failed to open DB");
-	  }
-	  return (-1);
+	    if (imspd_debug) {
+		fprintf(stderr,"%s: ", c->db);
+		perror("failed to open DB");
+	    }
+	    return (-1);
 	}
 	c->fd = fd;
     }
-
+    
 /* ESYS DOC - this is a very expensive proposition if the file is large.  For
    a short while, we will have allocated in virtual memory TWICE the size
    of the file.   This should be changed to some sort of chained buffer
@@ -486,52 +487,52 @@ static int loadcache(c, flags)
 /* : allocate a buffer to hold the database file text */
     data = (char *) malloc(stbuf.st_size + 1);
     if (data == NULL) {
-      if (imspd_debug) {
-	perror("malloc of data failed");
-      } else {
-	syslog(LOG_ERR, "imspd: Unable to allocate %d bytes for %s: %m",
-	       stbuf.st_size+1, c->db);
-      }
-      CLEANUP_RETURN(-1);
+	if (imspd_debug) {
+	    perror("malloc of data failed");
+	} else {
+	    syslog(LOG_ERR, "imspd: Unable to allocate %d bytes for %s: %m",
+		   stbuf.st_size+1, c->db);
+	}
+	CLEANUP_RETURN(-1);
     }
-
+    
 /* : read data from file into the data buffer */
     count = read(fd, data, stbuf.st_size);
     if (count != stbuf.st_size) {
-      if (imspd_debug) {
-	fprintf(stderr, "internal consistency error count(%d) != stbuf.st_size (%d)\n",
-		count, stbuf.st_size);
-      }
+	if (imspd_debug) {
+	    fprintf(stderr, "internal consistency error count(%d) != stbuf.st_size (%d)\n",
+		    count, stbuf.st_size);
+	}
 	CLEANUP_RETURN(-1);
     }
     data[count] = '\0';			/* set a sentinel -- THIS IS IMPORTANT!*/
-
+    
 /* : parse the database data into the cache */
     if (parsecache(c, data, flags) < 0) {
 	freecache(c);
 	if (imspd_debug) {
-	  fprintf(stderr,"parsecache failed\n");
+	    fprintf(stderr,"parsecache failed\n");
 	}
 	CLEANUP_RETURN(-1);
     }
-
+    
 /* : mark the cache as loaded and not modified */
     c->loaded = 1;
     c->modified = 0;
     c->icase = flags & SDB_ICASE;
-
+    
  CLEANUP:
 /* : free the database data buffer */
     if (data != NULL) {
 	free(data);
     }
-
+    
 /* : close the database file if unlocked */
     if (c->locks == 0) {
 	close(fd);
 	c->fd = -1;
     }
-
+    
 /* : return the return value */
     return(rtval);
 }
@@ -549,20 +550,43 @@ static int loadcache(c, flags)
 static int writecache(c)
     cache *c;
 {
+    int oldfd;
+    cache oldc;
     FILE *out;
-    int i;
-    char *scan;
+    int i, old = 0;
+    char *scan, *value;
     char newname[MAXDBPATHLEN + 5];
+    
+    memset(&oldc, 0, sizeof(cache));
 
 /* : open new database file for output */
     snprintf(newname, sizeof(newname), newext, c->db);
     if ((out = fopen(newname, "w")) == NULL) {
 	return (-1);
     }
-
+    oldfd = open(c->db, O_RDWR, 0600);
+    if (oldfd >= 0) { 
+        close(oldfd);
+	strcpy(oldc.db, c->db);
+	oldc.locks = 0;
+	if ((old = loadcache(&oldc, c->icase?SDB_ICASE:0)) < 0) 
+	    old = 0;
+	else
+	    old = 1;
+    }
+    
 /* : walk the cache writing keyvalue elements to the database file */
     for (i = 0; i < c->cachecount; ++i) {
 	if (c->kv[i].key == NULL) continue;
+	if ((c->kv[i].modified == 0) && old) {
+	    sdb_keyvalue *kv;
+	    
+	    kv = kv_bsearch(c->kv[i].key, oldc.kv, oldc.cachecount,
+			    c->icase ? strcasecmp : strcmp);
+	    value = kv ? kv->value : NULL;
+	} else {
+	    value = c->kv[i].value;
+	}
 	for (scan = c->kv[i].key; *scan; ++scan) {
 	    if (*scan == '\n') {
 		putc('\\', out);
@@ -577,9 +601,9 @@ static int writecache(c)
 		putc(*scan, out);
 	    }
 	}
-	if (c->kv[i].value != NULL) {
+	if (value != NULL) {
 	    putc(' ', out);
-	    for (scan = c->kv[i].value; *scan; ++scan) {
+	    for (scan = value; *scan; ++scan) {
 		if (*scan == '\n') {
 		    putc('\\', out);
 		    putc('n', out);
@@ -593,20 +617,28 @@ static int writecache(c)
 	}
 	putc('\n', out);
     }
-
+    
+    if (old) {
+	lock_unlock(oldc.fd);
+	close(oldc.fd);
+	oldc.fd = -1;
+	oldc.locks = 0;
+	freecache(&oldc);
+    }
+    
 /* : make sure write & rename succeed */
     if (fclose(out) == EOF || rename(newname, c->db) < 0) {
 	unlink(newname);
 	return (-1);
     }
-
+    
 /* : remove any stray locks */
     if (c->locks) {
 	lock_unlock(c->fd);
 	close(c->fd);
 	c->locks = -1;
     }
-
+    
 /* : return success */
     return (0);
 }
@@ -619,7 +651,7 @@ int sdb_init()
     int i;
     struct stat stbuf;
     char path[MAXPATHLEN];
-
+    
     /* initialize cache */
     for (i = 0; i < NUMGDBSTR; ++i) {
 	memset(globdb[i].db, '\0', sizeof (globdb[i].db));
@@ -638,7 +670,7 @@ int sdb_init()
 	privdb[i].cachesize = 0;
 	privdb[i].cachecount = 0;
     }
-
+    
     /* initialize directories */
     snprintf(path, sizeof(path), "%s/%s", PREFIX, PRIVPREFIX);
     if (stat(path, &stbuf) < 0) {
@@ -647,7 +679,7 @@ int sdb_init()
 	    return (-1);
 	}
     }
-
+    
     return (0);
 }
 
@@ -662,7 +694,7 @@ void sdb_done()
 {
     int i;
     cache *c;
-
+    
     /* write and free caches */
     for (i = 0; i < NUMGDBSTR; ++i) {
 	c = globdb + i;
@@ -697,28 +729,28 @@ void sdb_done()
  */
 
 void sdb_flush(flags)
-int flags;
+    int flags;
 {
     int i;
     cache *c;
-
+    
     /* write and free global caches (to /var/imsp/<db>) */
     if (flags & SDB_FLUSH_GLOBAL)
-      for (i = 0; i < NUMGDBSTR; ++i) {
-	c = globdb + i;
-	if (c->modified) {
-	    writecache(c);
+	for (i = 0; i < NUMGDBSTR; ++i) {
+	    c = globdb + i;
+	    if (c->modified) {
+		writecache(c);
+	    }
 	}
-      }
-
+    
     /* write and free private caches (to /var/imsp/user/.../<db>) */
     if (flags & SDB_FLUSH_PRIVATE) {
-      for (i = 0; i < NUMPDBSTR; ++i) {
-	c = privdb + i;
-	if (c->modified) {
-	    writecache(c);
+	for (i = 0; i < NUMPDBSTR; ++i) {
+	    c = privdb + i;
+	    if (c->modified) {
+		writecache(c);
+	    }
 	}
-      }
     }
 }
 
@@ -746,12 +778,12 @@ int sdb_create(db)
     char *scan;
     cache *c;
     struct stat stbuf;
-
+    
     /* verify that db is valid & doesn't exist */
     if ((c = findcache(db)) == NULL || c->loaded || stat(c->db, &stbuf) >= 0) {
 	return (-1);
     }
-
+    
     /* make user's subdirectory if needed */
     if (!strncmp(db, PRIVPREFIX, PRIVPREFIXLEN)) {
 	scan = strchr(c->db + PREFIXLEN + PRIVPREFIXLEN + 2, '/');
@@ -760,12 +792,12 @@ int sdb_create(db)
 	mkdir(c->db, 0700);
 	*scan = '/';
     }
-
+    
     /* create datafile */
     if ((fd = open(c->db, O_WRONLY | O_CREAT, 0600)) < 0 || close(fd) < 0) {
 	return (-1);
     }
-
+    
     return (0);
 }
 
@@ -779,18 +811,18 @@ int sdb_delete(db)
     cache *c;
     
     if ((c = findcache(db)) == NULL || c->locks) return (-1);
-
+    
     /* remove file and empty cache */
     if (unlink(c->db) < 0) return (-1);
     freecache(c);
-
+    
     /* try removing user directory for cleanliness sake */
     if (!strncmp(db, PRIVPREFIX, PRIVPREFIXLEN)) {
 	strcpy(lname, c->db);
 	*strrchr(lname, '/') = '\0';
 	rmdir(lname);
     }
-
+    
     return (0);
 }
 
@@ -805,49 +837,49 @@ int sdb_copy(dbsrc, dbdst, flags)
     cache *citem;
     char dbname[MAXDBPATHLEN+1];
     int fd=0, result;
-
+    
     /* create the destination. this locks and prevents another
-    * rename from happening. */
+     * rename from happening. */
     if (sdb_create(dbdst) < 0) {
-      return (-1);
+	return (-1);
     }
-
+    
     if ((citem = findcache(dbdst)) == NULL) {
-      sdb_delete(dbdst);
-      return (-1);
+	sdb_delete(dbdst);
+	return (-1);
     }
     strcpy(dbname, citem->db);
     /* grab lock on destination */
     if ((fd = open(dbname, O_RDWR|O_CREAT, 0600)) < 0) {
-      sdb_delete(dbdst);
-      return (-1);
+	sdb_delete(dbdst);
+	return (-1);
     }
-
+    
     if (lock_reopen(fd, citem->db, NULL, NULL) < 0) {
-      sdb_delete(dbdst);
-      close(fd);
-      return (-1);
+	sdb_delete(dbdst);
+	close(fd);
+	return (-1);
     }
     /* make sure files are valid */
     if (((citem = findcache(dbsrc)) == NULL) || citem->locks) {
-      sdb_delete(dbdst);
-      return (-1);
+	sdb_delete(dbdst);
+	return (-1);
     }
-
+    
     /* read in source */
     if (loadcache(citem, flags) < 0) {
 	lock_unlock(fd);
 	close(fd);
 	return (-1);
     }
-
+    
     /* change cache to be the destination */
     citem->fd = fd;
     strcpy(citem->db, dbname);
-
+    
     /* write the cache & unlock file */
     result = writecache(citem);
-
+    
     return (result);
 }
 
@@ -863,7 +895,7 @@ int sdb_get(db, key, flags, value)
 {
     cache *c;
     sdb_keyvalue *kv;
-
+    
     /* get db in cache */
     c = findcache(db);
     if (c == NULL) {
@@ -874,18 +906,18 @@ int sdb_get(db, key, flags, value)
 	    return (-1);
 	}
     }
-
+    
     /* if db empty, return no value */
     if (!c->cachecount) {
 	*value = NULL;
 	return (0);
     }
-
+    
     /* do binary search */
     kv = kv_bsearch(key, c->kv, c->cachecount,
 		    (flags & SDB_ICASE) ? strcasecmp : strcmp);
     *value = kv ? kv->value : NULL;
-
+    
     return (0);
 }
 
@@ -897,7 +929,7 @@ int sdb_count(db, flags)
     int flags;
 {
     cache *c;
-
+    
     /* get db in cache */
     c = findcache(db);
     if (c == NULL) {
@@ -908,7 +940,7 @@ int sdb_count(db, flags)
 	    return (-1);
 	}
     }
-
+    
     return (c->cachecount);
 }
 
@@ -926,7 +958,7 @@ static int valuematch(vpat, value)
     if ((vg = glob_init(vpat, 0)) == NULL) return (-1);
     result = GLOB_TEST(vg, value);
     glob_free(&vg);
-
+    
     return (result >= 0 ? 0 : 1);
 }
 
@@ -959,7 +991,7 @@ int sdb_match(db, key, flags, vpat, copy, pkv, count)
     sdb_keyvalue *ksrc, *kdst;
     glob *g, *vg;
     int gcount, copysize;
-
+    
     /* initialization */
     *pkv = NULL;
     *count = 0;
@@ -977,43 +1009,33 @@ int sdb_match(db, key, flags, vpat, copy, pkv, count)
 	    return (-1);
 	}
     }
-
+    
     /* if db empty, return no match */
     if (!c->cachecount) {
 	return (0);
     }
     
-   /* special case for no wildcards */
+    /* special case for no wildcards */
     for (scan = key; *scan != '*' && *scan != '%' && *scan != '?' && *scan;
 	 ++scan);
     if (!*scan) {
-
+	
 	/* do binary search */
 	ksrc = kv_bsearch(key, c->kv, c->cachecount,
 			  (flags & SDB_ICASE) ? strcasecmp : strcmp);
 	if (ksrc && valuematch(vpat, value = ksrc->value) == 0) {
 	    key = ksrc->key;
-	    kdst = *pkv = (sdb_keyvalue *)
-		malloc(sizeof (sdb_keyvalue)
-		       + (copy ? strlen(key) + strlen(value) + 2 : 0));
-	    if (kdst == NULL) return (-1);
-	    kdst->key = key;
-	    kdst->value = value;
+	    kdst = *pkv = (sdb_keyvalue *)xmalloc(sizeof (sdb_keyvalue));
+	    kdst->key = copy ? (char *)xstrdup(key) : key;
+	    kdst->value = copy ? (char *)xstrdup(value) : value;
 	    *count = 1;
-	    if (copy) {
-		scan = (char *) (kdst + 1);
-		kdst->key = scan;
-		while (*scan++ = *key++);
-		kdst->value = scan;
-		while (*scan++ = *value++);
-	    }
 	}
 	return (0);
     }
-
+    
     /* set key to NULL if it's a "*" */
     if (key[0] == '*' && key[1] == '\0') key = NULL;
-
+    
     /* make space for a complete match -- we can reduce usage later */
     kdst = *pkv = (sdb_keyvalue *) malloc(sizeof (sdb_keyvalue) * c->cachecount);
     if (kdst == NULL) {
@@ -1021,7 +1043,7 @@ int sdb_match(db, key, flags, vpat, copy, pkv, count)
     }
     memset(kdst, '\0', (sizeof(sdb_keyvalue) * c->cachecount));
     ksrc = c->kv;
-
+    
     /* special case for full match */
     if (!key && !vpat) {
 	memcpy((void *) kdst, (void *) ksrc, c->cachecount * sizeof (sdb_keyvalue));
@@ -1061,7 +1083,7 @@ int sdb_match(db, key, flags, vpat, copy, pkv, count)
 	}
 	if (vpat) glob_free(&vg);
 	if (key) glob_free(&g);
-
+	
 	/* check for no match */
 	if (kdst == *pkv) {
 	    free((char *) kdst);
@@ -1070,15 +1092,15 @@ int sdb_match(db, key, flags, vpat, copy, pkv, count)
 	    return (0);
 	}
     }
-
+    
     /* calculate the number of keyvalue pairs that matched */
     *count = kdst - *pkv;
-
+    
     /* adjust down amount of space used by the match array */
     if (*count < c->cachecount) {
 	*pkv = (sdb_keyvalue *) realloc((char *) *pkv, (*count * sizeof(sdb_keyvalue)));
     }
-
+    
     /* if copy requested then walk the match array duplicating contents */
     if (copy) {
 	kdst = *pkv + *count;
@@ -1090,7 +1112,7 @@ int sdb_match(db, key, flags, vpat, copy, pkv, count)
 		*pkv = NULL;
 		return(-1);
 	    }
-
+	    
 	    if(ksrc->value) {
 		value = ksrc->value;
 		ksrc->value = strdup(value);
@@ -1103,7 +1125,7 @@ int sdb_match(db, key, flags, vpat, copy, pkv, count)
 	    }
 	}
     }
-
+    
     /* cleanup and return success if there was a match */
     return (*pkv == NULL ? -1 : 0);
 }
@@ -1124,12 +1146,12 @@ void sdb_freematch(kv, count, copy)
     int copy;
 {
     int i;				/* loop counter */
-
+    
     /* sanity checks */
     if (kv == NULL) {
 	return;
     }
-
+    
     /* if the list is a copy then walk the kv list freeing keys and values */
     if (copy) {
 	for (i = 0; i < count; i++) {
@@ -1137,7 +1159,7 @@ void sdb_freematch(kv, count, copy)
 	    if (kv[i].value != NULL) free(kv[i].value);
 	}
     }
-
+    
     /* free the keyvalue list */
     free((char *) kv);
 }
@@ -1157,17 +1179,17 @@ int sdb_unlock(db, key, flags)
     int flags;
 {
     cache *c;
-
+    
     /* CLAIM - to unlock a cache, it *must* have been locked and therefore loaded.
        If we look for the cache and it is not instantiated, then there is a big
        problem. */
     /* find the appropriate cache entry */
     c = findcache(db);
     if (c == NULL) return (-1);
-
+    
     /* if there are no locks then just return success */
     if (c->locks == 0) return (0);
-
+    
     /*XXX: should write the cache if it's been modified */
     /* 11/6/97 - well, the problem is that if you do, performance
      * really blows as say for option_set, you basically do a 
@@ -1175,13 +1197,13 @@ int sdb_unlock(db, key, flags)
      * to be written to disk each time -- probably the right behavior but
      * it makes writes really bad.  So we don't do it. 
      */
-
+    
     /* decrement the lock count on the cache and if 0, remove lock file */
     c->locks--;
     if (c->locks == 0) {
-      lock_unlock(c->fd);
-      close(c->fd);
-      c->fd = -1;
+	lock_unlock(c->fd);
+	close(c->fd);
+	c->fd = -1;
     }
     
     /* return success */
@@ -1204,46 +1226,46 @@ int sdb_writelock(db, key, flags)
     int flags;
 {
     cache *c;
-
+    
     /* find the appropriate cache entry */
     c = findcache(db);
     if (c == NULL) {
-      if (imspd_debug) {
-	fprintf(stderr,"failed to find cache\n");
-      }
-      return(-1);
+	if (imspd_debug) {
+	    fprintf(stderr,"failed to find cache\n");
+	}
+	return(-1);
     }
-
+    
     /* if the cache is already locked then increment it and return success */
     if (c->locks) {
 	++c->locks;
 	return (0);
     }
-
+    
     /* get exclusive lock on database file */
     if (c->fd == -1) {
-      c->fd = open(c->db, O_RDWR);
-      if (c->fd < 0) return (-1);
+	c->fd = open(c->db, O_RDWR);
+	if (c->fd < 0) return (-1);
     } 
     if (lock_reopen(c->fd, c->db, NULL, NULL) < 0) {
-      if (imspd_debug) {
-	fprintf(stderr,"failed to reopen\n");
-      }
-      close(c->fd);
-      c->fd = -1;
-      return (-1);
+	if (imspd_debug) {
+	    fprintf(stderr,"failed to reopen\n");
+	}
+	close(c->fd);
+	c->fd = -1;
+	return (-1);
     }
     /* load the cache if necessary */
     if (c->loaded == 0) {
 	if (loadcache(c, flags) < 0) {
-	  if (imspd_debug) {
-	    fprintf(stderr,"failed to load cache\n");
-	  }
+	    if (imspd_debug) {
+		fprintf(stderr,"failed to load cache\n");
+	    }
 	    return (-1);
 	}
     }
     ++c->locks;
-
+    
     /* return success */
     return (0);
 }
@@ -1269,10 +1291,10 @@ int sdb_set(db, key, flags, value)
     int top, bot, mid, cmp;
     sdb_keyvalue *kvtop, *kvmid;
     int (*cmpf)() = (flags & SDB_ICASE) ? strcasecmp : strcmp;
-
+    
     /* find the appropriate cache entry & make sure it's locked */
     if ((c = findcache(db)) == NULL || !c->locks) return (-1);
-
+    
     /* if db not empty, look for the key */
     mid = 0;
     if (c->cachecount) {
@@ -1287,17 +1309,18 @@ int sdb_set(db, key, flags, value)
 		bot = mid + 1;
 	    }
 	}
-
+	
 	/* if we matched then set the value in the cache */
 	if (!cmp) {
 	    if (c->kv[mid].value != NULL) free(c->kv[mid].value);
 	    c->kv[mid].value = strdup(value);
 	    c->modified = 1;
+	    c->kv[mid].modified = 1;
 	    return(0);
 	}
 	if (cmp > 0) ++mid;
     }
-
+    
     /* grow the cache if necessary */
     if ((c->cachecount + 1) > c->cachesize) {
 	size = ((c->cachesize + CACHE_INCREMENT) * sizeof(sdb_keyvalue));
@@ -1313,7 +1336,7 @@ int sdb_set(db, key, flags, value)
 	}
 	c->cachesize += CACHE_INCREMENT;
     }
-
+    
     /* instantiate the new keyvalue pair */
     c->cachecount++;
     kvtop = c->kv + (c->cachecount - 1);
@@ -1324,10 +1347,11 @@ int sdb_set(db, key, flags, value)
     }
     kvmid->key = strdup(key);
     kvmid->value = strdup(value);
-
+    kvmid->modified = 1;
+    
     /* mark the cache as modified */
     c->modified = 1;
-
+    
     /* return success */
     return (0);
 }
@@ -1347,10 +1371,10 @@ int sdb_remove(db, key, flags)
 {
     cache *c;
     sdb_keyvalue *kvtop, *kvmid;
-
+    
     /* find the appropriate cache entry & make sure it's locked */
     if ((c = findcache(db)) == NULL || !c->locks) return (-1);
-
+    
     /* if db not empty, look for the key */
     if (!c->cachecount) return (-1);
     
@@ -1360,17 +1384,17 @@ int sdb_remove(db, key, flags)
     if (!kvmid) return (-1);
     if (kvmid->key != NULL) free(kvmid->key);
     if (kvmid->value != NULL) free(kvmid->value);
-
+    
     /* remove the key pair from the cache */
     kvtop = c->kv + --c->cachecount;
     while (kvmid < kvtop) {
 	kvmid[0] = kvmid[1];
 	++kvmid;
     }
-
+    
     /* mark the cache as modified */
     c->modified = 1;
-
+    
     /* return success */
     return (0);
 }
