@@ -1,7 +1,8 @@
 /* imclient.c -- Streaming IMxP client library
- $Id: imclient.c,v 1.5 2003/12/09 20:28:27 cdaboo Exp $
- 
- * Copyright (c) 1998-2000 Carnegie Mellon University.  All rights reserved.
+ *
+ * $Id: imclient.c,v 1.6 2004/01/22 18:01:28 rjs3 Exp $
+ *
+ * Copyright (c) 1998-2003 Carnegie Mellon University.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -42,6 +43,7 @@
  *
  */
 #include <config.h>
+#include <assert.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <ctype.h>
@@ -50,13 +52,12 @@
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
-#ifdef __STDC__
+#ifdef HAVE_STDARG_H
 #include <stdarg.h>
 #else
 #include <varargs.h>
 #endif
 #include <sys/types.h>
-#include <sys/time.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <netdb.h>
@@ -64,7 +65,7 @@
 #include <sys/select.h>
 #endif
 
-#include <sasl.h>
+#include <sasl/sasl.h>
 
 #ifdef HAVE_SSL
 #include <openssl/lhash.h>
@@ -80,8 +81,8 @@
 #include "imparse.h"
 #include "imclient.h"
 #include "nonblock.h"
-
-extern int errno;
+#include "util.h"
+#include "iptostring.h"
 
 /* I/O buffer size */
 #define IMCLIENT_BUFSIZE 4096
@@ -89,8 +90,8 @@ extern int errno;
 /* Command completion callback record */
 struct imclient_cmdcallback {
     struct imclient_cmdcallback *next;
-    long tag;			/* Command tag # */
-    imclient_proc_t *proc;		/* Callback function */
+    unsigned long tag;		/* Command tag # */
+    imclient_proc_t *proc;	/* Callback function */
     void *rock;			/* Callback rock */
 };
 
@@ -102,6 +103,12 @@ struct imclient_callback {
     void *rock;			/* Callback rock */
 };
 
+struct stringlist 
+{
+    char *str;
+    struct stringlist *next;
+};
+
 /* Connection data */
 struct imclient {
     /* TCP stream */
@@ -113,18 +120,18 @@ struct imclient {
     /* Data to be output to server */
     char outbuf[IMCLIENT_BUFSIZE];
     char *outptr;
-    int outleft;
+    size_t outleft;
     char *outstart;
 
     /* Replies being received from server */
     char *replybuf;
     char *replystart;
-    int replyliteralleft;
-    int replylen;
-    int alloc_replybuf;
+    size_t replyliteralleft;
+    size_t replylen;
+    size_t alloc_replybuf;
     
     /* Protection mechanism data */
-  /*    struct sasl_client *mech;
+    /* struct sasl_client *mech;
     sasl_encodefunc_t *encodefunc;
     sasl_decodefunc_t *decodefunc;*/
     void *state;
@@ -145,13 +152,15 @@ struct imclient {
     int callback_alloc;
     struct imclient_callback *callback;
 
-  sasl_conn_t *saslconn;
-  int saslcompleted;
+    struct stringlist *interact_results;
+
+    sasl_conn_t *saslconn;
+    int saslcompleted;
 
 #ifdef HAVE_SSL
-  SSL_CTX *tls_ctx;
-  SSL *tls_conn;
-  int tls_on; /* wheather we are under a layer or not */
+    SSL_CTX *tls_ctx;
+    SSL *tls_conn;
+    int tls_on; /* wheather we are under a layer or not */
 #endif /* HAVE_SSL */
 };
 
@@ -159,7 +168,7 @@ struct imclient {
  * Syntactic class of a character
  * 0 - literal, 1 - quoted-string, 2 - atom
  */
-static char charclass[256] = {
+static const char charclass[256] = {
     0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 1, 0, 1, 1, /*  00 -  0f */
     1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, /*  10 -  1f */
     1, 2, 0, 2, 2, 1, 2, 2, 1, 1, 1, 2, 2, 2, 2, 2, /* ' ' -  '/' */
@@ -183,27 +192,21 @@ static struct imclient_cmdcallback *cmdcallback_freelist;
 
 /* Forward declarations */
 void imclient_write(struct imclient *imclient,
-		    const char *s, unsigned len);
+		    const char *s, size_t len);
 static int imclient_writeastring P((struct imclient *imclient,
 				     const char *str));
 static void imclient_writebase64 P((struct imclient *imclient,
-				    const char *output, unsigned len));
+				    const char *output, size_t len));
 static void imclient_eof P((struct imclient *imclient));
 static int imclient_decodebase64 P((char *input));
 
 /* callbacks we support */
-static sasl_callback_t callbacks[] = {
-  {
-    SASL_CB_USER, NULL, NULL
-  }, {
-    SASL_CB_GETREALM, NULL, NULL
-  }, {
-    SASL_CB_AUTHNAME, NULL, NULL
-  }, {
-    SASL_CB_PASS, NULL, NULL    
-  }, {
-    SASL_CB_LIST_END, NULL, NULL
-  }
+static const sasl_callback_t callbacks[] = {
+  { SASL_CB_USER, NULL, NULL }, 
+  { SASL_CB_GETREALM, NULL, NULL }, 
+  { SASL_CB_AUTHNAME, NULL, NULL }, 
+  { SASL_CB_PASS, NULL, NULL },
+  { SASL_CB_LIST_END, NULL, NULL }
 };
 
 /*
@@ -219,44 +222,44 @@ int imclient_connect(struct imclient **imclient,
 		     const char *port, 
 		     sasl_callback_t *cbs)
 {
-    int s;
-    struct hostent *hp;
-    struct servent *sp;
-    struct sockaddr_in addr;
-    static struct imclient zeroimclient;
+    int s = -1;
+    struct addrinfo hints, *res0 = NULL, *res;
     int saslresult;
+    static int didinit;
 
-    hp = gethostbyname(host);
-    if (!hp) return -1;
+    assert(imclient);
+    assert(host);
 
-    s = socket(AF_INET, SOCK_STREAM, 0);
-    if (s == -1) return errno;
-
-    addr.sin_family = AF_INET;
-    memcpy(&addr.sin_addr, hp->h_addr, sizeof(addr.sin_addr));
-    if (port && imparse_isnumber(port)) {
-	addr.sin_port = htons(atoi(port));
+    if (!port)
+	port = "143";
+    memset(&hints, 0, sizeof(hints));
+    hints.ai_family = PF_UNSPEC;
+    hints.ai_socktype = SOCK_STREAM;
+    hints.ai_flags = AI_CANONNAME;
+    if (getaddrinfo(host, port, &hints, &res0))
+	return -1;
+    for (res = res0; res; res = res->ai_next) {
+	s = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
+	if (s < 0)
+	    continue;
+	if (connect(s, res->ai_addr, res->ai_addrlen) >= 0)
+	    break;
+	close(s);
+	s = -1;
     }
-    else if (port) {
-	sp = getservbyname(port, "tcp");
-	if (!sp) return -2;
-	addr.sin_port = sp->s_port;
-    }
-    else {
-	addr.sin_port = htons(143);
-    }
-    if (connect(s, (struct sockaddr *)&addr, sizeof(addr)) == -1) {
+    if (s < 0)
 	return errno;
-    }
     /*    nonblock(s, 1); */
-    *imclient = (struct imclient *)xmalloc(sizeof(struct imclient));
-    **imclient = zeroimclient;
+    *imclient = (struct imclient *)xzmalloc(sizeof(struct imclient));
     (*imclient)->fd = s;
     (*imclient)->saslconn = NULL;
     (*imclient)->saslcompleted = 0;
-    (*imclient)->servername = xstrdup(hp->h_name);
+    (*imclient)->servername = xstrdup(res0->ai_canonname ?
+				      res0->ai_canonname : host);
+    freeaddrinfo(res0);
     (*imclient)->outptr = (*imclient)->outstart = (*imclient)->outbuf;
     (*imclient)->outleft = (*imclient)->maxplain = sizeof((*imclient)->outbuf);
+    (*imclient)->interact_results = NULL;
     imclient_addcallback(*imclient,
 		 "", 0, (imclient_proc_t *) 0, (void *)0,
 		 "OK", CALLBACK_NOLITERAL, (imclient_proc_t *)0, (void *)0,
@@ -271,19 +274,18 @@ int imclient_connect(struct imclient **imclient,
     (*imclient)->tls_on=0;
 #endif /* HAVE_SSL */
 
-
-  /* attempt to start sasl */
-    if (cbs) {
-	saslresult=sasl_client_init(cbs);	
-    } else {
-	saslresult=sasl_client_init(callbacks);	
+    if (!didinit) {
+	/* attempt to start sasl */
+	saslresult = sasl_client_init(NULL);
+	if (saslresult!=SASL_OK) return 1;
+	didinit = 1;
     }
-    if (saslresult!=SASL_OK) return 1;
 
   /* client new connection */
   saslresult=sasl_client_new("imap", /* xxx ideally this should be configurable */
 			     (*imclient)->servername,
-			     NULL,
+			     NULL, NULL,
+			     cbs ? cbs : callbacks,
 			     0,
 			     &((*imclient)->saslconn));
   if (saslresult!=SASL_OK) return 1;
@@ -295,48 +297,53 @@ int imclient_connect(struct imclient **imclient,
  * Close and free the connection 'imclient'
  */
 void
-imclient_close(imclient)
-struct imclient *imclient;
+imclient_close(struct imclient *imclient)
 {
     int i;
+    struct stringlist *cur, *cur_next;
+
+    assert(imclient);
 
     imclient_eof(imclient);
     close(imclient->fd);
     free(imclient->servername);
     if (imclient->replybuf) free(imclient->replybuf);
     /*    if (imclient->state) imclient->mech->free_state(imclient->state);*/
-    
+    sasl_dispose(&(imclient->saslconn));
     for (i = 0; i < imclient->callback_num; i++) {
 	free(imclient->callback[i].keyword);
     }
     if (imclient->callback) free((char *)imclient->callback);
+
+    for(cur=imclient->interact_results; cur; cur=cur_next) {
+	cur_next = cur->next;
+	free(cur->str);
+	free(cur);
+    }
+
     free((char *)imclient);
 }
 
-void 
-imclient_setflags(imclient, flags)
-struct imclient *imclient;
-int flags;
+void imclient_setflags(struct imclient *imclient, int flags)
 {
+    assert(imclient);
     imclient->flags |= flags;
 }
 
-void
-imclient_clearflags(imclient, flags)
-struct imclient *imclient;
-int flags;
+void imclient_clearflags(struct imclient *imclient, int flags)
 {
+    assert(imclient);
     imclient->flags &= ~flags;
 }
 
 char *
-imclient_servername(imclient)
-struct imclient *imclient;
+imclient_servername(struct imclient *imclient)
 {
+    assert(imclient);
     return imclient->servername;
 }
 
-#define CALLBACKGROW 2 /* XXX 30 */
+#define CALLBACKGROW 5
 
 /*
  * Add untagged data callbacks to a connection.
@@ -378,6 +385,8 @@ va_dcl
     imclient = va_arg(pvar, struct imclient *);
 #endif
 
+    assert(imclient);
+
     while ((keyword = va_arg(pvar, char *))) {
 	flags = va_arg(pvar, int);
 	proc = va_arg(pvar, imclient_proc_t *);
@@ -465,6 +474,8 @@ va_dcl
     fmt = va_arg(pvar, char *);
 #endif
 
+    assert(imclient);
+
     imclient->gensym++;
     if (imclient->gensym <= 0) imclient->gensym = 1;
 
@@ -557,16 +568,16 @@ fail:
     }
 }
 
-static int
-imclient_writeastring(imclient, str)
-struct imclient *imclient;
-const char *str;
+static int imclient_writeastring(struct imclient *imclient, const char *str)
 {
     const char *p;
     unsigned len = 0;
     int class = 2;
     char buf[30];
 
+    assert(imclient);
+    assert(str);
+    
     for (p = str; *p; p++) {
 	len++;
 	if (class > charclass[(unsigned char)*p]) {
@@ -607,12 +618,11 @@ const char *str;
 /*
  * Write to the connection 'imclient' the data 's', of length 'len'
  */
-void
-imclient_write(imclient, s, len)
-struct imclient *imclient;
-const char *s;
-unsigned len;
+void imclient_write(struct imclient *imclient, const char *s, size_t len)
 {
+    assert(imclient);
+    assert(s);
+    
     /* If no data pending for output, reset the buffer */
     if (imclient->outptr == imclient->outstart) {
 	imclient->outstart = imclient->outptr = imclient->outbuf;
@@ -650,25 +660,23 @@ unsigned len;
  */
 #define REPLYSLACK 80		/* When growing, allocate this extra slack */
 #define REPLYSHRINK (4096+500)	/* If more than this free, shrink buffer */
-static void
-imclient_input(imclient, buf, len)
-struct imclient *imclient;
-char *buf;
-int len;
+static void imclient_input(struct imclient *imclient, char *buf, int len)
 {
-    long replytag;
+    unsigned long replytag;
     struct imclient_reply reply;
     char *endreply;
     char *p;
-    int parsed;
-    int literallen;
-    int keywordlen;
+    size_t parsed;
+    size_t literallen;
+    size_t keywordlen;
     int keywordindex;
     struct imclient_cmdcallback **cmdcb, *cmdcbtemp;
-    char *plainbuf;
-    int plainlen;
-    int freeplain;
+    const char *plainbuf;
+    unsigned plainlen;
     int result;
+
+    assert(imclient);
+    assert(buf);
     
     if (imclient->saslcompleted == 1) {
 	/* decrypt what we have */
@@ -678,11 +686,9 @@ int len;
 	}
 
 	if (plainlen == 0) return;
-	freeplain = 1;
     } else {
 	plainbuf = buf;
 	plainlen = len;
-	freeplain = 0;
     }
 
     /* Ensure replybuf has enough space to take the input */
@@ -715,31 +721,29 @@ int len;
 	}
     }
     
-
     /* Remember where new data starts */
     parsed = imclient->replylen;
 
     /* Copy the data to the buffer and NUL-terminate it */
     memcpy(imclient->replybuf + imclient->replylen, plainbuf, plainlen);
     imclient->replylen += plainlen;
-
-    if (freeplain && plainlen) {
-	free(plainbuf);
-    }
+    imclient->replybuf[imclient->replylen] = '\0';
 
     /* Process the new data (of length 'plainlen') */
     while (parsed < imclient->replylen) {
 	/* If we're reading a literal, skip over it. */
 	if (imclient->replyliteralleft) {
-	    if (plainlen > imclient->replyliteralleft) {
-		plainlen -= imclient->replyliteralleft;
+	    size_t avail;
+
+	    avail = imclient->replylen - parsed;
+
+	    if (avail > imclient->replyliteralleft) {
 		parsed += imclient->replyliteralleft;
 		imclient->replyliteralleft = 0;
 		continue;
-	    }
-	    else {
-		parsed += plainlen;
-		imclient->replyliteralleft -= plainlen;
+	    } else {
+		parsed += avail;
+		imclient->replyliteralleft -= avail;
 		return;
 	    }
 	}
@@ -753,7 +757,6 @@ int len;
 
 	parsed = endreply - imclient->replybuf + 1;
 
-
 	/* parse tag */
 	p = imclient->replystart;
 	if (*p == '+' && p[1] == ' ') {
@@ -776,7 +779,9 @@ int len;
 	}
 	else {
 	    replytag = 0;
-	    while (isdigit((int) *p)) replytag = replytag * 10 + *p++ - '0';
+	    while (isdigit((unsigned char) *p)) {
+		replytag = replytag * 10 + *p++ - '0';
+	    }
 	    if (*p++ != ' ') {
 		/* XXX Got junk from the server */
 		/* Start parsing the next reply */
@@ -786,9 +791,11 @@ int len;
 	}
 
 	/* parse num, if there */
-	if (replytag == 0 && isdigit((int) *p)) {
+	if (replytag == 0 && isdigit((unsigned char) *p)) {
 	    reply.msgno = 0;
-	    while (isdigit((int) *p)) reply.msgno = reply.msgno * 10 + *p++ - '0';
+	    while (isdigit((unsigned char) *p)) {
+		reply.msgno = reply.msgno * 10 + *p++ - '0';
+	    }
 	    if (*p++ != ' ') {
 		/* XXX Got junk from the server */
 		/* Start parsing the next reply */
@@ -802,7 +809,7 @@ int len;
 	
 	/* parse keyword */
 	reply.keyword = p;
-	while (*p != ' ' && *p != '\n') p++;
+	while (*p && *p != ' ' && *p != '\n') p++;
 	keywordlen = p - reply.keyword;
 	reply.text = p + 1;
 	if (*p == '\n') {
@@ -824,17 +831,23 @@ int len;
 
 
 	    /* Scan back and see if the end of the line introduces a literal */
-	    if (!iscompletion && endreply[-1] == '\r' && endreply[-2] == '}' &&
-		isdigit((int) endreply[-3])) {
+	    if (!iscompletion && endreply > imclient->replystart+2 &&
+		endreply[-1] == '\r' && endreply[-2] == '}' &&
+		isdigit((unsigned char) endreply[-3])) {
 		p = endreply - 4;
-		while (p > imclient->replystart && isdigit((int) *p)) p--;
+		while (p > imclient->replystart && 
+		       isdigit((unsigned char) *p)) {
+		    p--;
+		}
 		if (p > imclient->replystart + 2 && *p == '{' &&
 		    charclass[(unsigned char)p[-1]] != 2) {
 
 		    /* Parse the size of the literal */
 		    literallen = 0;
 		    p++;
-		    while (isdigit((int) *p)) literallen = literallen*10 + *p++ -'0';
+		    while (isdigit((unsigned char) *p)) {
+		        literallen = literallen*10 + *p++ -'0';
+		    }
 
 		    /* Do a continue to read literal & following line */
 		    imclient->replyliteralleft = literallen;
@@ -892,17 +905,23 @@ int len;
 
 	/* Scan back and see if the end of the line introduces a literal */
 	if (!(imclient->callback[keywordindex].flags & CALLBACK_NOLITERAL)) {
-	    if (endreply[-1] == '\r' && endreply[-2] == '}' &&
-		isdigit((int) endreply[-3])) {
+	    if (endreply > imclient->replystart+2 &&
+		endreply[-1] == '\r' && endreply[-2] == '}' &&
+		isdigit((unsigned char) endreply[-3])) {
 		p = endreply - 4;
-		while (p > imclient->replystart && isdigit((int) *p)) p--;
+		while (p > imclient->replystart && 
+		       isdigit((unsigned char) *p)) {
+		    p--;
+		}
 		if (p > imclient->replystart + 2 && *p == '{' &&
 		    charclass[(unsigned char)p[-1]] != 2) {
 
 		    /* Parse the size of the literal */
 		    literallen = 0;
 		    p++;
-		    while (isdigit((int) *p)) literallen = literallen*10 + *p++ -'0';
+		    while (isdigit((unsigned char) *p)) {
+		        literallen = literallen*10 + *p++ -'0';
+		    }
 
 		    /* Do a continue to read literal & following line */
 		    imclient->replyliteralleft = literallen;
@@ -928,13 +947,13 @@ int len;
  * Received an EOF on the connection 'imclient'
  * Issue appropriate callbacks.
  */
-static void
-imclient_eof(imclient)
-struct imclient *imclient;
+static void imclient_eof(struct imclient *imclient)
 {
     struct imclient_cmdcallback *cmdcb;
     struct imclient_reply reply;
 
+    assert(imclient);
+
     imclient->readytag = 0;
     imclient->readytxt = 0;
 
@@ -960,12 +979,13 @@ struct imclient *imclient;
  * 'wanttowrite' is filled in with nonzero value iff should
  * select() for write as well.
  */
-void
-imclient_getselectinfo(imclient, fd, wanttowrite)
-struct imclient *imclient;
-int *fd;
-int *wanttowrite;
+void imclient_getselectinfo(struct imclient *imclient, int *fd,
+			    int *wanttowrite)
 {
+    assert(imclient);
+    assert(fd);
+    assert(wanttowrite);
+    
     *fd = imclient->fd;
     *wanttowrite = imclient->outptr - imclient->outstart;
 }
@@ -973,23 +993,23 @@ int *wanttowrite;
 /*
  * Process one input or output event on the connection 'imclient'.
  */
-void
-imclient_processoneevent(imclient)
-struct imclient *imclient;
+void imclient_processoneevent(struct imclient *imclient)
 {
-    char buf[4+IMCLIENT_BUFSIZE];
+    char buf[IMCLIENT_BUFSIZE];
     int n;
     int writelen;
     fd_set rfds, wfds;
     FD_ZERO(&rfds);
     FD_ZERO(&wfds);
 
+    assert(imclient);
+
     for (;;) {
 	writelen = imclient->outptr - imclient->outstart;
 
 	if ((imclient->saslcompleted==1) && (writelen>0)) {
 	    unsigned int cryptlen=0;
-	    char *cryptptr=NULL;
+	    const char *cryptptr=NULL;
 
 	  if (sasl_encode(imclient->saslconn, imclient->outstart, writelen,
 			  &cryptptr,&cryptlen)!=SASL_OK)
@@ -1011,12 +1031,10 @@ struct imclient *imclient;
 #endif /* HAVE_SSL */
 	  	  
 	  if (n > 0) {	    
-	    free(cryptptr);
 	    imclient->outstart += writelen;
 	    return;
 	  }
 
-
 	  /* XXX Also EPIPE & the like? */
 	  /* Make sure we select() for writing */
 
@@ -1088,14 +1106,15 @@ struct authresult {
 };
 
 /* Command completion callback for imclient_authenticate */
-static void 
-authresult(imclient, rock, reply)
-struct imclient *imclient;
-void *rock;
-struct imclient_reply *reply;
+static void authresult(struct imclient *imclient __attribute__((unused)),
+		       void *rock,
+		       struct imclient_reply *reply)
 {
     struct authresult *result = (struct authresult *)rock;
 
+    assert(result);
+    assert(reply);
+
     if (!strcmp(reply->keyword, "OK")) {
 	result->replytype = replytype_ok;
     }
@@ -1106,16 +1125,17 @@ struct imclient_reply *reply;
 }
 
 /* Command completion for starttls */
-static void 
-tlsresult(imclient, rock, reply)
-struct imclient *imclient;
-void *rock;
-struct imclient_reply *reply;
+static void tlsresult(struct imclient *imclient __attribute__((unused)),
+		      void *rock,
+		      struct imclient_reply *reply)
 {
     struct authresult *result = (struct authresult *)rock;
 
+    assert(result);
+    assert(reply);
+
     if (!strcmp(reply->keyword, "OK")) {
-      result->replytype = replytype_ok;
+	result->replytype = replytype_ok;
     }
     else if (!strcmp(reply->keyword, "NO")) {
 	result->replytype = replytype_no;
@@ -1126,57 +1146,71 @@ struct imclient_reply *reply;
 
 static sasl_security_properties_t *make_secprops(int min,int max)
 {
-  sasl_security_properties_t *ret=(sasl_security_properties_t *)
-  malloc(sizeof(sasl_security_properties_t));
+  sasl_security_properties_t *ret=
+      (sasl_security_properties_t *)xzmalloc(sizeof(sasl_security_properties_t));
 
-  ret->maxbufsize=1024;
-  ret->min_ssf=min;
-  ret->max_ssf=max;
-
-  ret->security_flags=0;
-  ret->property_names=NULL;
-  ret->property_values=NULL;
+  ret->maxbufsize = IMCLIENT_BUFSIZE;
+  ret->min_ssf = min;
+  ret->max_ssf = max;
 
   return ret;
 }
 
-void interaction (sasl_interact_t *t, char *user)
+void interaction (struct imclient *context, sasl_interact_t *t, char *user)
 {
   char result[1024];
+  struct stringlist *cur;
+  
+  assert(context);
+  assert(t);
+
+  cur = malloc(sizeof(struct stringlist));
+  if(!cur) {
+      t->len=0;
+      t->result=NULL;
+      return;
+  }
+
+  cur->str = NULL;
+  cur->next = context->interact_results;
+  context->interact_results = cur;
 
   if ((t->id == SASL_CB_USER || t->id == SASL_CB_AUTHNAME) 
             && user && user[0]) {
       t->len = strlen(user);
-      t->result = xstrdup(user);
+      cur->str = xstrdup(user);
   } else {
       printf("%s: ", t->prompt);
       if (t->id == SASL_CB_PASS) {
 	  char *ptr = getpass("");
-	  strncpy(result, ptr, sizeof(result));
+	  strlcpy(result, ptr, sizeof(result));
       } else {
-	  fgets(result, sizeof(result), stdin);
+	  fgets(result, sizeof(result)-1, stdin);
 	  result[strlen(result) - 1] = '\0';
       }
 
       t->len = strlen(result);
-      t->result = (char *) xmalloc(t->len+1);
-      memset(t->result, 0, t->len+1);
-      memcpy((char *) t->result, result, t->len);
+      cur->str = (char *) xmalloc(t->len+1);
+      memset(cur->str, 0, t->len+1);
+      memcpy(cur->str, result, t->len);
   }
+
+  t->result = cur->str;
 }
 
-void fillin_interactions(sasl_interact_t *tlist, char *user)
+void fillin_interactions(struct imclient *context,
+			 sasl_interact_t *tlist, char *user)
 {
-  while (tlist->id!=SASL_CB_LIST_END)
-  {
-    interaction(tlist, user);
-    tlist++;
-  }
+    assert(context);
+    assert(tlist);
 
+    while (tlist->id!=SASL_CB_LIST_END)
+    {
+	interaction(context, tlist, user);
+	tlist++;
+    }
 }
 
-
-
 /*
  * Params:
  *  mechlist: list of mechanisms seperated by spaces
@@ -1186,10 +1220,8 @@ void fillin_interactions(sasl_interact_t *tlist, char *user)
  *  1 - failure
  *  2 - severe failure?
  */
-
 static int imclient_authenticate_sub(struct imclient *imclient, 
 				     char *mechlist, 
-				     char *service, 
 				     char *user,
 				     int minssf, 
 				     int maxssf,
@@ -1197,21 +1229,22 @@ static int imclient_authenticate_sub(struct imclient *imclient,
 {
   int saslresult;
   sasl_security_properties_t *secprops=NULL;
-  int addrsize=sizeof(struct sockaddr_in);
-  struct sockaddr_in *saddr_l=malloc(sizeof(struct sockaddr_in));
-  struct sockaddr_in *saddr_r=malloc(sizeof(struct sockaddr_in));
+  socklen_t addrsize;
+  struct sockaddr_storage saddr_l;
+  struct sockaddr_storage saddr_r;
+  char localip[60], remoteip[60];
   sasl_interact_t *client_interact=NULL;
-  char *out;
+  const char *out;
   unsigned int outlen;
   int inlen;
   struct authresult result;
 
-
-
+  assert(imclient);
+  assert(mechlist);
+  
   /*******
    * Now set the SASL properties
    *******/
-
   secprops=make_secprops(minssf,maxssf);
   if (secprops==NULL) return 1;
 
@@ -1219,24 +1252,28 @@ static int imclient_authenticate_sub(struct imclient *imclient,
   if (saslresult!=SASL_OK) return 1;
   free(secprops);
 
-  if (getpeername(imclient->fd,(struct sockaddr *)saddr_r,&addrsize)!=0)
-    return 1;
+  addrsize=sizeof(struct sockaddr_storage);
+  if (getpeername(imclient->fd,(struct sockaddr *)&saddr_r,&addrsize)!=0)
+      return 1;
 
-  /*  saddr_r->sin_port=htons(saddr_r->sin_port);	*/
-  saslresult=sasl_setprop(imclient->saslconn, SASL_IP_REMOTE, saddr_r);
+  addrsize=sizeof(struct sockaddr_storage);
+  if (getsockname(imclient->fd,(struct sockaddr *)&saddr_l,&addrsize)!=0)
+      return 1;
+
+  if(iptostring((const struct sockaddr *)&saddr_l, addrsize,
+		localip, sizeof(localip)) != 0)
+      return 1;
+
+  if(iptostring((const struct sockaddr *)&saddr_r, addrsize,
+		remoteip, sizeof(remoteip)) != 0)
+      return 1;
+
+  saslresult=sasl_setprop(imclient->saslconn, SASL_IPREMOTEPORT, remoteip);
   if (saslresult!=SASL_OK) return 1;
 
-  addrsize=sizeof(struct sockaddr_in);
-  if (getsockname(imclient->fd,(struct sockaddr *)saddr_l,&addrsize)!=0)
-    return 1;
-
-  /*  saddr_l->sin_port=htons(saddr_l->sin_port);	*/
-  saslresult=sasl_setprop(imclient->saslconn,   SASL_IP_LOCAL, saddr_l);
+  saslresult=sasl_setprop(imclient->saslconn, SASL_IPLOCALPORT, localip);
   if (saslresult!=SASL_OK) return 1;
 
-  free(saddr_l);
-  free(saddr_r);
-
   /********
    * SASL is setup. Now try the actual authentication
    ********/
@@ -1247,11 +1284,12 @@ static int imclient_authenticate_sub(struct imclient *imclient,
   while (saslresult==SASL_INTERACT)
   {
     saslresult=sasl_client_start(imclient->saslconn, mechlist,
-				 NULL, &client_interact,
+				 &client_interact,
 				 &out, &outlen,
 				 mechusing);
     if (saslresult==SASL_INTERACT) {
-	fillin_interactions(client_interact, user); /* fill in prompts */
+	fillin_interactions(imclient,
+			    client_interact, user); /* fill in prompts */
     }
   }
 
@@ -1270,7 +1308,7 @@ static int imclient_authenticate_sub(struct imclient *imclient,
     /* stop looping on command completion */
     if (!imclient->readytxt) break;
 
-    if (isspace((int) *imclient->readytxt)) {
+    if (isspace((unsigned char) *imclient->readytxt)) {
 	inlen = 0;
     } else {
 	inlen = imclient_decodebase64(imclient->readytxt);
@@ -1296,78 +1334,98 @@ static int imclient_authenticate_sub(struct imclient *imclient,
 	    
 	    if (saslresult == SASL_INTERACT) {
 		/* fill in prompts */
-		fillin_interactions(client_interact, user); 
+		fillin_interactions(imclient,
+				    client_interact, user); 
 	    }
 	}
     }
 
-    /* send to server */
-    /* Send our reply to the server */
+    /* send our reply to the server */
     if ((saslresult==SASL_OK) || (saslresult==SASL_CONTINUE)) {
-	imclient_writebase64(imclient, out, outlen);
+        if (out == NULL || outlen == 0) {
+            imclient_write(imclient, "\r\n", 2);
+        } else {
+            imclient_writebase64(imclient, out, outlen);
+        }
     } else {
-	imclient_write(imclient,"*\r\n",3);
+	imclient_write(imclient,"*\r\n", 3);
 	return saslresult;
     }
 
-    if (outlen > 0) { 
-	free(out); 
-    }
     outlen = 0;
   }
 
-  imclient->saslcompleted = 1;
+  if(result.replytype == replytype_ok) imclient->saslcompleted = 1;
 
   return (result.replytype != replytype_ok);
 }
 
-int 
-imclient_authenticate(struct imclient *imclient, 
-		      char *mechlist, 
-		      char *service, 
-		      char *user,
-		      int minssf, 
-		      int maxssf)
+/* xxx service is not needed here */
+int imclient_authenticate(struct imclient *imclient, 
+			  char *mechlist, 
+			  char *service __attribute__((unused)),
+			  char *user,
+			  int minssf, 
+			  int maxssf)
 {
     int r;
     char *mlist;
     const char *mtried;
 
+    assert(imclient);
+    assert(mechlist);
+
     mlist = xstrdup(mechlist);
+    ucase(mlist);
 
     do {
-	char *newlist;
-	char *tmp;
-	
 	mtried = NULL;
 
 	r = imclient_authenticate_sub(imclient,
 				      mlist,
-				      service,
 				      user,
 				      minssf,
 				      maxssf,
 				      &mtried);
 
 	/* eliminate mtried (mechanism tried) from mlist */
-	if (mtried) {
-	    newlist = xmalloc(strlen(mlist)+1);
-	    
-	    tmp = strstr(mlist,mtried);
+	if (r != 0 && mtried) {
+	    char *newlist = xmalloc(strlen(mlist)+1);
+	    char *mtr = xstrdup(mtried);
+	    char *tmp;
+
+	    ucase(mtr);
+	    tmp = strstr(mlist,mtr);
+	    if(!tmp) {
+		free(mtr);
+		free(mlist);
+		break;
+	    }
 	    *tmp = '\0';
 	    strcpy(newlist,mlist);
 	    
-	    tmp = strchr(tmp,' ');
+	    /* Use tmp+1 here to skip the \0 we just put in.
+	     * this is safe because even if the mechs are one character
+	     * long there would still be another trailing \0 */
+	    tmp = strchr(tmp+1,' ');
 	    if (tmp) {		
-		tmp++;
+		tmp++; /* skip the space */
 		strcat(newlist,tmp);
 	    }
 
+	    free(mtr);
 	    free(mlist);
 	    mlist = newlist;
 	}
     } while ((r != 0) && (mtried));
 
+    if (r == 0) {
+	const int *ptr;
+
+	sasl_getprop(imclient->saslconn, SASL_MAXOUTBUF, (const void **) &ptr);
+	imclient->maxplain = *ptr < IMCLIENT_BUFSIZE ? *ptr : IMCLIENT_BUFSIZE;
+    }
+
     free(mlist);
 
     return r;
@@ -1410,6 +1468,8 @@ static int imclient_decodebase64(char *input)
     unsigned char *output = (unsigned char *)input;
     int c1, c2, c3, c4;
 
+    assert(input);
+
     while (*input) {
 	c1 = *input++;
 	if (CHAR64(c1) == XX) return -1;
@@ -1436,18 +1496,19 @@ static int imclient_decodebase64(char *input)
  * Write to the connection 'imclient' the base-64 encoded data
  * 'output', of (unencoded) length 'len'.
  */
-static void
-imclient_writebase64(imclient, output, len)
-struct imclient *imclient;
-const char *output;
-unsigned len;
+static void imclient_writebase64(struct imclient *imclient,
+				 const char *output,
+				 size_t len)
 {
     char buf[1024];
-    int buflen = 0;
+    size_t buflen = 0;
     int c1, c2, c3;
 
+    assert(imclient);
+    assert(output);
+
     while (len) {
-	if (buflen >= sizeof(buf)-4) {
+	if (buflen >= (size_t)(sizeof(buf)-4)) {
 	    imclient_write(imclient, buf, buflen);
 	    buflen = 0;
 	}
@@ -1493,7 +1554,6 @@ unsigned len;
 
 static int verify_depth;
 static int verify_error = X509_V_OK;
-static int do_dump = 1;
 
 #define CCERT_BUFSIZ 256
 static char peer_CN[CCERT_BUFSIZ];
@@ -1545,7 +1605,7 @@ static int verify_callback(int ok, X509_STORE_CTX * ctx)
     err = X509_STORE_CTX_get_error(ctx);
     depth = X509_STORE_CTX_get_error_depth(ctx);
 
-    X509_NAME_oneline(X509_get_subject_name(err_cert), buf, 256);
+    X509_NAME_oneline(X509_get_subject_name(err_cert), buf, sizeof(buf));
 
     /*    if (verbose==1)
 	  printf("Peer cert verify depth=%d %s\n", depth, buf);*/
@@ -1563,7 +1623,8 @@ static int verify_callback(int ok, X509_STORE_CTX * ctx)
     }
     switch (ctx->error) {
     case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT:
-	X509_NAME_oneline(X509_get_issuer_name(ctx->current_cert), buf, 256);
+	X509_NAME_oneline(X509_get_issuer_name(ctx->current_cert),
+			  buf, sizeof(buf));
 	printf("issuer= %s\n", buf);
 	break;
     case X509_V_ERR_CERT_NOT_YET_VALID:
@@ -1584,8 +1645,9 @@ static int verify_callback(int ok, X509_STORE_CTX * ctx)
 
 
 /* taken from OpenSSL apps/s_cb.c */
-
-static RSA *tmp_rsa_cb(SSL * s, int export, int keylength)
+static RSA *tmp_rsa_cb(SSL *s __attribute__((unused)),
+		       int export __attribute__((unused)),
+		       int keylength)
 {
     static RSA *rsa_tmp = NULL;
 
@@ -1595,46 +1657,19 @@ static RSA *tmp_rsa_cb(SSL * s, int export, int keylength)
     return (rsa_tmp);
 }
 
-/* taken from OpenSSL apps/s_cb.c 
- * tim - this seems to just be giving logging messages
+/*
+ * Seed the random number generator.
  */
-
-static void apps_ssl_info_callback(SSL * s, int where, int ret)
+static int tls_rand_init(void)
 {
-    char   *str;
-    int     w;
-
-    /*    return; */ /* only useful for debugging */
-
-    w = where & ~SSL_ST_MASK;
-
-    if (w & SSL_ST_CONNECT)
-	str = "SSL_connect";
-    else if (w & SSL_ST_ACCEPT)
-	str = "SSL_accept";
-    else
-	str = "undefined";
-
-    if (where & SSL_CB_LOOP) {
-      printf("%s:%s\n", str, SSL_state_string_long(s));
-    } else if (where & SSL_CB_ALERT) {
-	str = (where & SSL_CB_READ) ? "read" : "write";
-	if ((ret & 0xff) != SSL3_AD_CLOSE_NOTIFY)
-	  printf("SSL3 alert %s:%s:%s\n", str,
-		   SSL_alert_type_string_long(ret),
-		   SSL_alert_desc_string_long(ret));
-    } else if (where & SSL_CB_EXIT) {
-	if (ret == 0)
-	    printf("%s:failed in %s\n",
-		     str, SSL_state_string_long(s));
-	else if (ret < 0) {
-	    printf("%s:error in %s %i\n",
-		     str, SSL_state_string_long(s),ret);
-	}
-    }
+#ifdef EGD_SOCKET
+    return (RAND_egd(EGD_SOCKET));
+#else
+    /* otherwise let OpenSSL do it internally */
+    return 0;
+#endif
 }
 
-
 char *var_tls_CAfile="";
 char *var_tls_CApath="";
  /*
@@ -1653,11 +1688,16 @@ static int tls_init_clientengine(struct imclient *imclient,
     char   *c_cert_file;
     char   *c_key_file;
 
+    assert(imclient);
 
     SSL_load_error_strings();
     SSLeay_add_ssl_algorithms();
+    if (tls_rand_init() == -1) {
+	printf("[ TLS engine: cannot seed PRNG ]\n");
+	return -1;
+    }
 
-    imclient->tls_ctx = SSL_CTX_new(SSLv23_client_method());
+    imclient->tls_ctx = SSL_CTX_new(TLSv1_client_method());
     if (imclient->tls_ctx == NULL) {
 	return -1;
     };
@@ -1704,6 +1744,9 @@ static int tls_init_clientengine(struct imclient *imclient,
     return 0;
 }
 
+#if 0 /* Dead code only for debugging */
+static int do_dump = 1;
+
 /*
  * taken from OpenSSL crypto/bio/b_dump.c, modified to save a lot of strcpy
  * and strcat by Matti Aarnio.
@@ -1738,7 +1781,7 @@ static int tls_dump(const char *s, int len)
 	buf[0] = '\0';				/* start with empty string */
 	ss = buf;
 
-	snprintf(ss, sizeof(buf), "%04x ", i * DUMP_WIDTH);
+	sprintf(ss, "%04x ", i * DUMP_WIDTH);
 	ss += strlen(ss);
 	for (j = 0; j < DUMP_WIDTH; j++) {
 	    if (((i * DUMP_WIDTH) + j) >= len) {
@@ -1746,7 +1789,7 @@ static int tls_dump(const char *s, int len)
 	    } else {
 		ch = ((unsigned char) *((char *) (s) + i * DUMP_WIDTH + j))
 		    & 0xff;
-		snprintf(ss, buf + sizeof(buf) - ss, "%02x[%c]%c", ch, ch, j == 7 ? '|' : ' ');
+		sprintf(ss, "%02x[%c]%c", ch, ch, j == 7 ? '|' : ' ');
 		ss += 6;
 	    }
 	}
@@ -1769,7 +1812,7 @@ static int tls_dump(const char *s, int len)
     }
 #ifdef TRUNCATE
     if (trunc > 0) {
-	snprintf(buf, sizeof(buf), "%04x - <SPACES/NULS>\n", len+ trunc);
+	sprintf(buf, "%04x - <SPACES/NULS>\n", len+ trunc);
 	printf("%s\n", buf);
 	ret += strlen(buf);
     }
@@ -1777,11 +1820,13 @@ static int tls_dump(const char *s, int len)
     return (ret);
 }
 
-
-/* taken from OpenSSL apps/s_cb.c */
-
-static long bio_dump_cb(BIO * bio, int cmd, const char *argp, int argi,
-			long argl, long ret)
+/* these next two taken from OpenSSL apps/s_cb.c */
+static long bio_dump_cb(BIO * bio,
+			int cmd,
+			const char *argp,
+			int argi,
+			long argl __attribute__((unused)),
+			long ret)
 {
     if (!do_dump)
 	return (ret);
@@ -1801,8 +1846,42 @@ static long bio_dump_cb(BIO * bio, int cmd, const char *argp, int argi,
     return (ret);
 }
 
+static void apps_ssl_info_callback(SSL * s, int where, int ret)
+{
+    char   *str;
+    int     w;
+
+    w = where & ~SSL_ST_MASK;
+
+    if (w & SSL_ST_CONNECT)
+	str = "SSL_connect";
+    else if (w & SSL_ST_ACCEPT)
+	str = "SSL_accept";
+    else
+	str = "undefined";
+
+    if (where & SSL_CB_LOOP) {
+      printf("%s:%s\n", str, SSL_state_string_long(s));
+    } else if (where & SSL_CB_ALERT) {
+	str = (where & SSL_CB_READ) ? "read" : "write";
+	if ((ret & 0xff) != SSL3_AD_CLOSE_NOTIFY)
+	    printf("SSL3 alert %s:%s:%s\n", str,
+		   SSL_alert_type_string_long(ret),
+		   SSL_alert_desc_string_long(ret));
+    } else if (where & SSL_CB_EXIT) {
+	if (ret == 0)
+	    printf("%s:failed in %s\n",
+		     str, SSL_state_string_long(s));
+	else if (ret < 0) {
+	    printf("%s:error in %s %i\n",
+		     str, SSL_state_string_long(s),ret);
+	}
+    }
+}
+#endif
+
 int tls_start_clienttls(struct imclient *imclient,
-			int *layer, char **authid, int fd)
+			unsigned *layer, char **authid, int fd)
 {
     int     sts;
     SSL_SESSION *session;
@@ -1896,16 +1975,16 @@ int tls_start_clienttls(struct imclient *imclient,
 }
 
 int imclient_starttls(struct imclient *imclient,
-			     int verifydepth,
+			     int verifydepth __attribute__((unused)),
 			     char *var_tls_cert_file, 
 			     char *var_tls_key_file,
-			     int *layer)
+			     int *layer __attribute__((unused)))
 {
   int result;
   struct authresult theresult;
-
-  sasl_external_properties_t externalprop;
-
+  unsigned ssf;
+  char *auth_id;
+  
   imclient_send(imclient, tlsresult, (void *)&theresult,
 		"STARTTLS");
 
@@ -1921,7 +2000,7 @@ int imclient_starttls(struct imclient *imclient,
     printf("[ TLS engine failed ]\n");
     return 1;
   } else {
-    result=tls_start_clienttls(imclient, &externalprop.ssf, &externalprop.auth_id, imclient->fd);
+    result=tls_start_clienttls(imclient, &ssf, &auth_id, imclient->fd);
     
     if (result!=0) {
       printf("[ TLS negotiation did not succeed ]\n");
@@ -1936,14 +2015,18 @@ int imclient_starttls(struct imclient *imclient,
 
   imclient->tls_on = 1;
 
-  externalprop.auth_id=""; /* xxx this really should be peer_CN or
-  issuer_CN but I can't figure out which is which at the moment */
+  auth_id=""; /* xxx this really should be peer_CN or
+		 issuer_CN but I can't figure out which is
+		 which at the moment */
 
   /* tell SASL about the negotiated layer */
   result=sasl_setprop(imclient->saslconn,
 		      SASL_SSF_EXTERNAL,
-		      &externalprop);
-  
+		      &ssf);
+  if (result!=SASL_OK) return 1;
+  result=sasl_setprop(imclient->saslconn,
+		      SASL_AUTH_EXTERNAL,
+		      auth_id);
   if (result!=SASL_OK) return 1;
 
   return 0;
