@@ -3,7 +3,7 @@
  * This file is like map_shared.c, but doesn't use any "SLOP", since
  * "Digital Unix" doesn't permit mapping a region past the end of a file.
  
- $Id: map_stupidshared.c,v 1.2 2000/12/08 03:24:43 jackson Exp $
+ $Id: map_stupidshared.c,v 1.3 2003/12/09 20:28:27 cdaboo Exp $
  
  * Copyright (c) 1998-2000 Carnegie Mellon University.  All rights reserved.
  *
@@ -75,7 +75,7 @@ const char *mboxname;
 	if (fstat(fd, &sbuf) == -1) {
 	    syslog(LOG_ERR, "IOERROR: fstating %s file%s%s: %m", name,
 		   mboxname ? " for " : "", mboxname ? mboxname : "");
-	    sprintf(buf, "failed to fstat %s file", name);
+	    snprintf(buf, sizeof(buf), "failed to fstat %s file", name);
 	    fatal(buf, EC_IOERR);
 	}
 	newlen = sbuf.st_size;
@@ -97,7 +97,7 @@ const char *mboxname;
     if (*base == (char *)-1) {
 	syslog(LOG_ERR, "IOERROR: mapping %s file%s%s: %m", name,
 	       mboxname ? " for " : "", mboxname ? mboxname : "");
-	sprintf(buf, "failed to mmap %s file", name);
+	snprintf(buf, sizeof(buf), "failed to mmap %s file", name);
 	fatal(buf, EC_IOERR);
     }
     *len = newlen;
