@@ -41,7 +41,7 @@
  *
  */
 /*
- * $Id: prot.c,v 1.5 2003/01/20 19:17:41 rjs3 Exp $
+ * $Id: prot.c,v 1.6 2003/12/09 20:28:27 cdaboo Exp $
  */
 
 #include <config.h>
@@ -470,7 +470,7 @@ int prot_fill(struct protstream *s)
 
 		if (s->log_timeptr) {
 		    time(&newtime);
-		    sprintf(timebuf, "<%ld<", newtime - *s->log_timeptr);
+		    snprintf(timebuf, sizeof(timebuf), "<%ld<", newtime - *s->log_timeptr);
 		    write(s->logfd, timebuf, strlen(timebuf));
 		    *s->log_timeptr = newtime;
 		}
@@ -520,7 +520,7 @@ int prot_flush(struct protstream *s)
 
 	if (s->log_timeptr) {
 	    time(&newtime);
-	    sprintf(timebuf, ">%ld>", newtime - *s->log_timeptr);
+	    snprintf(timebuf, sizeof(timebuf), ">%ld>", newtime - *s->log_timeptr);
 	    write(s->logfd, timebuf, strlen(timebuf));
 	}
 
@@ -635,13 +635,13 @@ int prot_printf(struct protstream *s, const char *fmt, ...)
 
 	case 'd':
 	    i = va_arg(pvar, int);
-	    sprintf(buf, "%d", i);
+	    snprintf(buf, sizeof(buf), "%d", i);
 	    prot_write(s, buf, strlen(buf));
 	    break;
 
 	case 'u':
 	    u = va_arg(pvar, int);
-	    sprintf(buf, "%u", u);
+	    snprintf(buf, sizeof(buf), "%u", u);
 	    prot_write(s, buf, strlen(buf));
 	    break;
 
